import rdflib

from typing import Union, Optional, List

from cas.linkml_ops.schema import (
    convert_cas_schema_to_linkml,
    decorate_linkml_schema,
    expand_schema,
)
from cas.linkml_ops.data import dump_to_rdf, populate_ids


def export_to_rdf(
    cas_schema: Optional[Union[str, dict]],
    data: Union[str, dict],
    ontology_namespace: str,
    ontology_iri: str,
    labelsets: Optional[List[str]] = None,
    output_path: str = None,
    validate: bool = True,
    include_cells: bool = True,
) -> rdflib.Graph:
    """
    Generates and returns an RDF graph from the provided data and CAS schema, with an option to write the RDF graph to a file.
    Args:
        cas_schema (Optional[Union[str, dict]]):
            Name of the CAS release (such as `base`, `cap`, `bican`), path to the CAS schema file, or CAS schema JSON object.
            If not provided, reads the `base` CAS schema from the CAS module.
        data (Union[str, dict]):
            The data JSON file path or JSON object dictionary.
        ontology_namespace (str):
            Ontology namespace (e.g., `MTG`).
        ontology_iri (str):
            Ontology IRI (e.g., `https://purl.brain-bican.org/ontology/AIT_MTG/`).
        labelsets (Optional[List[str]]):
            Labelsets used in the taxonomy, such as `["Cluster", "Subclass", "Class"]`.
        output_path (Optional[str]):
            Path to the output RDF file, if specified.
        validate (bool):
            Determines if data-schema validation checks will be performed. True by default.
        include_cells (bool):
            Determines if cell data will be included in the RDF output. True by default.
    Returns:
        An RDFlib graph object.
    """
    # Prepare the linkml schema
    base_linkml_schema = convert_cas_schema_to_linkml(cas_schema)
    decorated_schema = decorate_linkml_schema(
        base_linkml_schema,
        ontology_namespace=ontology_namespace,
        ontology_iri=ontology_iri,
        labelsets=labelsets,
    )
    expanded_schema = expand_schema(
        config=None, yaml_obj=decorated_schema, value_set_names=["CellTypeEnum"]
    )

    # Prepare the data
    instance = populate_ids(
        data,
        ontology_namespace=ontology_namespace,
        ontology_id=ontology_namespace,
    )
    rdf_graph = dump_to_rdf(
        schema=expanded_schema,
        instance=instance,
        ontology_namespace=ontology_namespace,
        ontology_iri=ontology_iri,
        labelsets=labelsets,
        validate=validate,
        include_cells=include_cells,
        output_path=output_path,
    )

    return rdf_graph



#!/usr/bin/env python3

"""
Script Description:
This script processes and integrates information from a JSON file and an AnnData (Annotated Data) file,
creating a new AnnData object that incorporates the metadata. The resulting AnnData object is then saved to a new file.

Key Features:
1. Parses command-line arguments for input JSON file, input AnnData file, and output file.
2. Reads and processes the input JSON file and AnnData file.
3. Updates the AnnData object with information from the JSON annotations and root keys.
4. Writes the modified AnnData object to a specified output file.
"""
import shutil

import h5py
import numpy as np
import pandas as pd

from cas.file_utils import read_json_file, update_obs_dataset, write_json_to_hdf5

LABELSET_NAME = "name"

LABELSET = "labelset"

LABELSETS = "labelsets"

ANNOTATIONS = "annotations"

CELL_IDS = "cell_ids"

CELL_LABEL = "cell_label"


def is_list_of_strings(var):
    """
    Check if a value is a list of strings.

    Parameters:
        var (list or any): The value to be checked.

    Returns:
        bool: True if the value is a list containing only string elements,
              False otherwise.

    """
    return isinstance(var, list) and all(isinstance(item, str) for item in var)


def flatten(json_file_path, anndata_file_path, output_file_path):
    """
     Processes and integrates information from a JSON file and an AnnData (Annotated Data) file, creating a new AnnData
     object that incorporates the metadata. The resulting AnnData object is then saved to a new file.

    Args:
        json_file_path: The path to the CAS json file.
        anndata_file_path: The path to the AnnData file.
        output_file_path: Output AnnData file name.
    """
    input_json = read_json_file(json_file_path)

    flatten_cas_object(input_json, anndata_file_path, output_file_path)


# @profile
def flatten_cas_object(input_json, anndata_file_path, output_file_path):
    """
     Processes and integrates information from a JSON file and an AnnData (Annotated Data) file, creating a new AnnData
     object that incorporates the metadata. The resulting AnnData object is then saved to a new file.

    Args:
        input_json: CAS json file.
        anndata_file_path: The path to the AnnData file.
        output_file_path: Output AnnData file name.
    """
    if output_file_path:
        shutil.copy(anndata_file_path, output_file_path)
        anndata_file_path = output_file_path

    annotations = input_json[ANNOTATIONS]
    parent_cell_ids = collect_parent_cell_ids(input_json)

    with h5py.File(anndata_file_path, "r+") as f:
        obs_dataset = f["obs"]
        obs_index = np.array(obs_dataset[obs_dataset.attrs['_index']], dtype=str)

        # obs
        flatten_data = process_annotations(annotations, obs_index, parent_cell_ids)
        update_obs_dataset(obs_dataset, flatten_data)

        # uns
        uns_json = generate_uns_json(input_json)
        uns_dataset = f["uns"]
        write_json_to_hdf5(uns_dataset, uns_json)


def process_annotations(annotations, obs_index, parent_cell_ids):
    """
    Processes annotations and generates flattened data for obs dataset.

    Args:
        annotations (list): List of annotations.
        obs_index (np.ndarray): Array representing the index of the obs dataset.
        parent_cell_ids (dict): Dictionary containing parent cell ids.

    Returns:
        dict: Dictionary containing flattened data.
    """
    flatten_data = {}
    for ann in annotations:
        cell_ids = ann.get(
            CELL_IDS, parent_cell_ids.get(ann.get("cell_set_accession", []))
        )
        if not cell_ids:  # only happens if data has multi-inheritance (as in basal ganglia data)
            continue
        # Convert cell_ids to a list if it's not already for np.isin
        if not isinstance(cell_ids, list):
            cell_ids = list(cell_ids)
        mask = np.isin(obs_index, cell_ids)

        for k, v in ann.items():
            if k in [CELL_IDS, LABELSET]:
                continue
                
            key = f"{ann[LABELSET]}--{k}" if k != CELL_LABEL else ann[LABELSET]
            value = ", ".join(
                sorted([str(value) for value in v] if isinstance(v, list) else [str(v)])
            )

            if key not in flatten_data:
                flatten_data[key] = pd.Series("", index=obs_index)
            flatten_data[key].loc[mask] = value

    return flatten_data


def generate_uns_json(input_json):
    """
    Generates a dictionary representing the uns (unstructured) field in an AnnData object from a given JSON input.

    This function processes information from a JSON input and generates a dictionary that represents the uns (unstructured)
    field in an AnnData object. The resulting dictionary can be used to populate the uns field in the AnnData object.

    Args:
        input_json (dict): A dictionary representing the input JSON data containing annotations.

    Returns:
        dict: A dictionary representing the uns (unstructured) field in an AnnData object, ready to be used as input
              for writing to an AnnData file.

    """
    uns_json = {}
    root_keys = list(input_json.keys())
    root_keys.remove(ANNOTATIONS)
    for key in root_keys:
        value = input_json[key]
        metadata_json = {}
        if is_list_of_strings(value):
            uns_json[key] = ", ".join(sorted(value))
        elif isinstance(value, str):
            uns_json[key] = value
        else:
            for labelset in value:
                metadata_key = labelset.get(LABELSET_NAME, '')
                metadata_json.update({metadata_key: {}})
                for k, v in labelset.items():
                    if k == LABELSET_NAME:
                        continue
                    metadata_json.get(metadata_key, {}).update({k: v})
        uns_json["cellannotation_metadata"] = metadata_json
    return uns_json


def collect_parent_cell_ids(cas):
    """
    Collects parent cell IDs from the given CAS (Cluster Annotation Service) data.

    This function iterates through labelsets in the CAS data and collects parent cell IDs
    associated with each labelset annotation. It populates and returns a dictionary
    mapping parent cell set accessions to sets of corresponding cell IDs.

    Args:
        cas (dict): The Cluster Annotation Service data containing labelsets and annotations.

    Returns:
        dict: A dictionary mapping parent cell set accessions to sets of corresponding cell IDs.
    """
    parent_cell_ids = dict()

    labelsets = sorted(cas[LABELSETS], key=lambda x: int(x["rank"]))
    for labelset in labelsets:
        ls_annotations = [
            ann for ann in cas[ANNOTATIONS] if ann["labelset"] == labelset["name"]
        ]

        for ann in ls_annotations:
            if "parent_cell_set_accession" in ann:
                cell_ids = set()
                if CELL_IDS in ann and ann[CELL_IDS]:
                    cell_ids = set(ann[CELL_IDS])
                elif (
                    "cell_set_accession" in ann
                    and ann["cell_set_accession"] in parent_cell_ids
                ):
                    cell_ids = parent_cell_ids[ann["cell_set_accession"]]

                if ann["parent_cell_set_accession"] in parent_cell_ids:
                    parent_cell_ids[ann["parent_cell_set_accession"]].update(cell_ids)
                else:
                    parent_cell_ids[ann["parent_cell_set_accession"]] = set(cell_ids)

    return parent_cell_ids



from cas.accession.base_accession_manager import BaseAccessionManager


class IncrementalAccessionManager(BaseAccessionManager):
    """
    Numerically incremental Accession ID generator.
    """

    def __init__(self, accession_prefix=None, last_accession_id=0):
        """
        Initializer.
        Params:
            accession_prefix: accession_id prefix
        """
        if accession_prefix is None:
            accession_prefix = ""
        self.accession_prefix = accession_prefix
        self.last_accession_id = last_accession_id
        self.accession_ids = list()

    def generate_accession_id(self, id_recommendation: str = None, labelset: str = None) -> str:
        """
        Generates an auto-increment based accession id. If the recommended accession_id is available, uses it.
        Params:
            id_recommendation: accession id recommendation. Function uses this id if it is available,
            provides an auto-incremented id otherwise.
            labelset: this parameter is not utilized in this implementation.
        Return: accession_id
        """
        if id_recommendation:
            id_recommendation = id_recommendation.replace(self.accession_prefix, "")
            if id_recommendation.startswith("_"):
                id_recommendation = id_recommendation[1:]

        if (
            id_recommendation
            and id_recommendation not in self.accession_ids
            and id_recommendation.isdigit()
            # and int(id_recommendation) > self.last_accession_id
            and int(id_recommendation) not in self.accession_ids
        ):
            accession_id = id_recommendation
            if self.last_accession_id < int(id_recommendation):
                self.last_accession_id = int(id_recommendation)
        elif id_recommendation and not id_recommendation.isdigit():
            # non-numeric accession id
            accession_id = id_recommendation
        else:
            id_candidate = self.last_accession_id + 1
            while str(id_candidate) in self.accession_ids:
                id_candidate += 1
            accession_id = str(id_candidate)
            self.last_accession_id = id_candidate

        self.accession_ids.append(accession_id)
        if self.accession_prefix and not accession_id.startswith(self.accession_prefix):
            accession_id = self.accession_prefix + accession_id

        return accession_id



import json
import logging
from typing import Any, Dict, List, Optional, Union

import numpy as np
import pandas as pd

from cas.file_utils import read_json_file


# Configure logging
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


def add_author_annotations(
    cas: Dict[str, Any],
    df: pd.DataFrame,
    join_column: Union[str, List[str]],
    columns: Optional[List[str]] = None,
) -> Dict[str, Any]:
    """
    Updates the provided CAS dictionary with author annotation fields from a DataFrame.
    Annotations are based on the values matched between the DataFrame's specified join columns and the CAS records.
    All columns or specified columns are added as annotations.

    Args:
        cas: The CAS dictionary loaded from a JSON file.
        df: DataFrame containing the data to join with the CAS.
        join_column: The column(s) in the DataFrame used for matching CAS records; can be a single column or a list of columns.
        columns: Optional list of columns whose data will be added as annotations. If None, all DataFrame columns are used.

    Returns:
        Dict[str, Any]: The CAS dictionary updated with new annotation fields.
    """
    validate_columns(df, join_column, columns)
    validate_values(df, join_column, cas)
    annotations = cas["annotations"]
    for annotation in annotations:
        labelset = annotation["labelset"]
        filter_value = (
            annotation[join_column]
            if isinstance(join_column, str)
            else annotation["cell_label"]
        )
        author_annotation_dict = dataframe_to_dict(
            df, join_column, labelset, filter_value, columns
        )
        if "author_annotation_fields" in annotation:
            annotation["author_annotation_fields"].update(author_annotation_dict)
        else:
            annotation["author_annotation_fields"] = author_annotation_dict

    return cas


def add_author_annotations_from_file(
    cas_json: str,
    csv_path: str,
    join_column: Union[str, List[str]],
    columns: Optional[List[str]] = None,
    output_file_path: str = "output.json",
) -> None:
    """
    Reads data from a CSV file and a CAS JSON file, then updates the CAS JSON with author annotation fields,
    and outputs the updated CAS JSON to a specified file. It uses specified columns or all columns if none are
    specified.

    Args:
        cas_json: Path to the CAS JSON file.
        csv_path: Path to the CSV file.
        join_column: Column name or names used for matching CAS records.
        columns: Optional columns to be added as annotations; if None, all columns are used.
        output_file_path: Output CAS file name.

    Returns:
        Dict[str, Any]: The updated CAS JSON dictionary.
    """
    cas = read_json_file(cas_json)
    df = pd.read_csv(csv_path)
    output_cas = add_author_annotations(cas, df, join_column, columns)
    # Write the JSON data to the file
    with open(output_file_path, "w") as json_file:
        json.dump(output_cas, json_file, indent=2)


def validate_columns(
    df: pd.DataFrame, join_column: Union[str, List[str]], columns: Optional[List[str]]
):
    """
    Validates the existence of key columns and optionally other specified columns in a DataFrame.

    Args:
        df: DataFrame to check.
        join_column: Column or columns that must exist in the DataFrame.
        columns: Additional column names that must exist if specified; checks only join_column if None.

    Raises:
        ValueError: If any specified columns, including the join column(s), are not found in the DataFrame.
    """
    # Normalize join_column to always be a list
    if isinstance(join_column, str):
        join_column = [join_column]
    column_list_to_check = join_column + (columns if columns else [])
    missing_columns = [col for col in column_list_to_check if col not in df.columns]
    if missing_columns:
        raise ValueError(f"Missing columns in DataFrame: {', '.join(missing_columns)}")


def validate_values(
    df: pd.DataFrame, join_column: Union[str, List[str]], cas: Dict[str, Any]
):
    """
    Validates that all keys generated from specified DataFrame columns exist within the CAS annotations.
    If there are any keys in the DataFrame that are not found in the CAS annotations, it raises an exception
    indicating these are extra, undesired keys.

    Parameters:
        df (pd.DataFrame): DataFrame containing the data to validate.
        join_column (Union[str, List[str]]): Column or list of columns in the DataFrame used to form the keys.
            If a list of two strings is provided, these columns are concatenated to form the keys.
        cas (Dict[str, Any]): Dictionary containing CAS annotations, each annotation expected to contain the keys.

    Raises:
        ValueError: If there are extra keys in the DataFrame that do not exist in the CAS annotations.
    """
    cas_key_list = []
    for annotation in cas["annotations"]:
        if isinstance(join_column, list) and len(join_column) == 2:
            key_value = annotation.get(join_column[0], "") + annotation.get(
                join_column[1], ""
            )
            cas_key_list.append(key_value)
        else:
            cas_key_list.append(annotation.get(join_column, ""))

    if isinstance(join_column, list) and len(join_column) == 2:
        df_key_list = df[join_column[0]].str.cat(df[join_column[1]], sep="").tolist()
    else:
        df_key_list = df[join_column].tolist()

    extra_keys = [key for key in df_key_list if key not in cas_key_list]
    empty_rows = [key for key in cas_key_list if key not in df_key_list]
    if extra_keys:
        raise ValueError(
            f"Extra keys in DataFrame that are not in CAS data: {extra_keys}"
        )
    if empty_rows:
        logging.info(f"Following values in {' '.join(join_column) if isinstance(join_column, list) else join_column} exist in "
                     f"CAS data but missing from DataFrame:"
                     f" {' '.join(empty_rows)}")


def dataframe_to_dict(
    df: pd.DataFrame,
    join_column: Union[str, List[str]],
    labelset: str,
    filter_value: str,
    columns: Optional[List[str]],
) -> dict:
    """
    Converts specified columns of a DataFrame into a dictionary. If columns are not specified, the entire DataFrame
    for rows matching the condition is converted.

    Args:
        df: DataFrame to extract data from.
        join_column: The column or columns used to filter the DataFrame based on cell_label.
        labelset: Labelset of a CAS annotation.
        filter_value: The value in join_column used to filter rows.
        columns: List of column names to convert to dictionary. If None, all columns are used.

    Returns:
        dict: Dictionary with column names as keys and single values or lists of values as values.
    """
    # Handle different cases for join_column being a string or list of strings
    if join_column == "cell_set_accession":
        filtered_df = df[df[join_column] == filter_value]
    elif isinstance(join_column, str):
        filtered_df = df[df[join_column] == filter_value]
    elif isinstance(join_column, list) and len(join_column) == 2:
        # Expect cell_label to be a tuple or list with corresponding values
        filtered_df = df[
            (df[join_column[0]] == labelset) & (df[join_column[1]] == filter_value)
        ]
    else:
        raise ValueError(
            "join_column must be either a string or a list of two strings."
        )

    # Determine columns to include in the output dictionary
    if columns is None:
        # Exclude join_column(s) from the output
        exclude_columns = (
            join_column if isinstance(join_column, list) else [join_column]
        )
        columns = [col for col in df.columns if col not in exclude_columns]

    # Select the specified columns to include
    filtered_df = filtered_df[columns]

    filtered_df.replace(' ', np.nan, inplace=True)
    filtered_df.replace(np.nan, None, inplace=True)
    result_dict = filtered_df.to_dict(orient="list")
    # Unpack lists that contain only a single item
    for key, value in result_dict.items():
        if len(value) == 1:
            result_dict[key] = value[0]

    return result_dict
